#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from sysrepo_interfaces import utils as utils
from sysrepo_interfaces import vpp as vpputils
from ucpeif import ipsecif as ips
from netaddr import IPNetwork, IPAddress

def ipsecvpn_disconnect(self, ipsvpn):
    ret = False
    if ipsvpn is None:
        return
    try:
        ret = ipsvpn.disconnect()
        if ret is False:
            self.logger.error('Disconnect IPsecVPN interface \'{}\' failed'.format(ipsvpn.name))

        ipsvpn.unload_connection()
        ipsvpn.unload_security()
    except:
        self.logger.error('Failed to disconnect IPsecVPN interface')

def ipsecvpn_connect(self, ipsvpn):
    ret = False
    if ipsvpn is None:
        return
    try:
        ipsvpn.load_security()
        ipsvpn.load_connection(debug=True)
        ret = ipsvpn.connect(timeout=10000)
        if ret is False:
            self.logger.error('Connect IPsecVPN interface \'{}\' failed'.format(ipsvpn.name))
    except:
        self.logger.error('Failed to connect IPsecVPN interface')

def set_ippass_ipsec(self, old_enable, new_enable):
    if old_enable == new_enable:
        return

    self.logger.info('IPSEC is enabled {} - set IPPass'.format(new_enable))

    vpputils.set_ippass_enable_udp_dst_port(self, 500, new_enable)
    vpputils.set_ippass_enable_udp_dst_port(self, 4500, new_enable)
    vpputils.set_ippass_enable_esp_proto(self, new_enable)

def config_ipsecvpn_startup(self, name, config, ipsvpn_interfaces):
    if not config:
        return

    for ipsvpn in ipsvpn_interfaces:
        if ipsvpn.name == name:
            for key, value in config.items():
                if key == "ike-version":
                    if value == '1':
                        ipsvpn.ike = ips.IKE.V1
                    elif value == '2':
                        ipsvpn.ike = ips.IKE.V2
                if key == "ike-mode":
                    if value == "main":
                        ipsvpn.ike1_mode = ips.IKE1_MODE.MAIN
                    elif value == "aggressive":
                        ipsvpn.ike1_mode = ips.IKE1_MODE.AGGRESIVE
                elif key == "ike-phase-1":
                    if 'dh-groups' in config[key]:
                        for group in config[key]['dh-groups']:
                            ipsvpn.dh_group_add(int(group['group']), ips.IKE_PHASE.PHASE1)
                    if 'proposals' in config[key]:
                        for prop in config[key]['proposals']:
                            if 'encryption-algorithm' in prop and 'authentication-algorithm' in prop:
                                ipsvpn.proposal_add(str(prop['encryption-algorithm']), str(prop['authentication-algorithm']), ips.IKE_PHASE.PHASE1)
                    if 'local' in config[key] and 'rounds' in config[key]['local']:
                        for auth_method in config[key]['local']['rounds']:
                            if 'authentication-method' in auth_method:
                                if auth_method['authentication-method'] == 'pre-shared-keys':
                                    ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.PSK, psk=auth_method['secret']))
                                elif auth_method['authentication-method'] == 'pubkeys':
                                    passphrase = None if 'key-pass' not in auth_method.keys() else str(auth_method['key-pass'])
                                    ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.PUBKEY, ca_crt=auth_method['ca-cert'], crt=auth_method['cert'], key=auth_method['key'], passphrase=passphrase))
                                elif auth_method['authentication-method'] == 'xauth':
                                    ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.XAUTH, user=auth_method['username'], psw=auth_method['password']))
                                elif auth_method['authentication-method'] == 'eap':
                                    passphrase = None if 'key-pass' not in auth_method.keys() else str(auth_method['key-pass'])
                                    ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.EAP, user=auth_method['username'], psw=auth_method['password'], ca_crt=auth_method['ca-cert'], crt=auth_method['cert'], key=auth_method['key'], passphrase=passphrase))
                    if 'remote' in config[key] and 'rounds' in config[key]['remote']:
                        for auth_method in config[key]['remote']['rounds']:
                            if 'authentication-method' in auth_method:
                                if auth_method['authentication-method'] == 'pre-shared-keys':
                                    ipsvpn.auth_add(ips.AUTH_SIDE.REMOTE, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.PSK, psk=auth_method['secret']))
                                elif auth_method['authentication-method'] == 'pubkeys':
                                    ipsvpn.auth_add(ips.AUTH_SIDE.REMOTE, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.PUBKEY, ca_crt=auth_method['ca-cert'], crt=auth_method['cert']))
                elif key == "ike-phase-2":
                    if 'dh-groups' in config[key]:
                        for group in config[key]['dh-groups']:
                            ipsvpn.dh_group_add(int(group['group']), ips.IKE_PHASE.PHASE2)
                    if 'proposals' in config[key]:
                        for prop in config[key]['proposals']:
                            if 'encryption-algorithm' in prop and 'authentication-algorithm' in prop:
                                ipsvpn.proposal_add(str(prop['encryption-algorithm']), str(prop['authentication-algorithm']), ips.IKE_PHASE.PHASE2)
                    if 'local-ts' in config[key] and config[key]['local-ts']:
                        ipsvpn.local_ts = IPNetwork(config[key]['local-ts'])
                    if 'remote-ts' in config[key] and config[key]['remote-ts']:
                        ipsvpn.remote_ts = IPNetwork(config[key]['remote-ts'])
                elif key == 'remote-gw':
                    ipsvpn.gateway = str(value)
                elif key == 'local-id':
                    ipsvpn.local_id = str(value)
                elif key == 'remote-id':
                    ipsvpn.remote_id = str(value)
                elif key == 'config-mode':
                    if value == 'pull':
                        ipsvpn.mode_config = ips.MODE_CONFIG.PULL
                    elif value == 'push':
                        ipsvpn.mode_config = ips.MODE_CONFIG.PUSH
                elif key == 'protocol-mode':
                    if value == 'tunnel':
                        ipsvpn.child_mode = ips.CHILD_MODE.TUNNEL
                    elif value == 'transport':
                        ipsvpn.child_mode = ips.CHILD_MODE.TRANSPORT
                elif key == 'install-virtual-ip':
                    if value is False or value == 'false':
                        ipsvpn.vips = None
                    else:
                        ipsvpn.vips = IPAddress(config.get('virtual-ip', '0.0.0.0'))

            if 'enable' in config and 'true' in str(config['enable']).lower():
                set_ippass_ipsec(self, False, config.get('enable', False))
                ipsecvpn_connect(self, ipsvpn)

def config_ipsecvpn_apply(self, name, change_set, ipsvpn_interfaces):
    if not change_set:
        return

    rounds = {'local': [{}, {}], 'remote': [{}]}
    props = {'ike-phase-1': [{}, {}, {}, {}, {}], 'ike-phase-2': [{}, {}, {}, {}, {}]}
    old_props = {'ike-phase-1': [{}, {}, {}, {}, {}], 'ike-phase-2': [{}, {}, {}, {}, {}]}
    l_rounds = self.get_ike1_local_rounds(name)
    if l_rounds:
        for l in l_rounds:
            rounds['local'][l['id'] - 1] = l

    r_rounds = self.get_ike1_remote_rounds(name)
    if r_rounds:
        for r in r_rounds:
            rounds['remote'][r['id'] - 1] = r

    for ipsvpn in ipsvpn_interfaces:
        if ipsvpn.name == name:
            if_status = self.get_interface_param(name, 'enable')
            inst_vip = self.get_install_vip(name)
            vip = self.get_vip(name)
            prev_status = if_status
            for xpath, change in change_set.items():
                # Get proposal and round IDs
                prop_id = change['xpath'][change['xpath'].find('proposals[id=\'')+len('proposals[id=\''):change['xpath'].rfind('\'')]
                round_id = change['xpath'][change['xpath'].find('rounds[id=\'')+len('rounds[id=\''):change['xpath'].rfind('\'')]

                if change['operation'] == 'DELETED':
                    if xpath.find('remote-gw') > -1:
                        ipsvpn.gateway = None
                    if xpath.find('local-id') > -1:
                        del ipsvpn.local_id
                    if xpath.find('remote-id') > -1:
                        del ipsvpn.remote_id
                    if xpath.find('ike-version') > -1:
                        ipsvpn.ike = ips.IKE.V2
                    if xpath.find('ike-mode') > -1:
                        ipsvpn.ike1_mode = ips.IKE1_MODE.MAIN
                    if xpath.find('rounds') > -1:
                        if xpath.find('local/rounds') > -1 and xpath.endswith('\']'):
                            rounds['local'][int(round_id) - 1] = {}
                            ipsvpn.auth_del(ips.AUTH_SIDE.LOCAL, int(round_id))
                        elif xpath.find('remote/rounds') > -1 and xpath.endswith('\']'):
                            rounds['remote'][int(round_id) - 1] = {}
                            ipsvpn.auth_del(ips.AUTH_SIDE.REMOTE, int(round_id))
                        elif xpath.find('local') > -1 and xpath.find('key-pass') > -1:
                            rounds['local'][int(round_id) - 1]['key-pass'] = None
                            rounds['local'][int(round_id) - 1]['modified'] = True
                    if xpath.find('ike-phase-1/dh-group') > -1:
                        if xpath.endswith(']') or not change['old_value']:
                            continue
                        dh_group = change['old_value']
                        if dh_group and dh_group !='None':
                            ipsvpn.dh_group_del(int(dh_group), ips.IKE_PHASE.PHASE1)
                    if xpath.find('ike-phase-2/dh-group') > -1:
                        if xpath.endswith(']') or not change['old_value']:
                            continue
                        dh_group = change['old_value']
                        if dh_group and dh_group !='None':
                            ipsvpn.dh_group_del(int(dh_group), ips.IKE_PHASE.PHASE2)
                    if xpath.find('ike-phase-1/proposals') > -1:
                        if xpath.find('encryption-algorithm') > -1:
                            old_props['ike-phase-1'][int(prop_id) - 1]['id'] = prop_id
                            old_props['ike-phase-1'][int(prop_id) - 1]['enc'] = change['old_value']
                        if xpath.find('authentication-algorithm') > -1:
                            old_props['ike-phase-1'][int(prop_id) - 1]['id'] = prop_id
                            old_props['ike-phase-1'][int(prop_id) - 1]['auth'] = change['old_value']
                    if xpath.find('ike-phase-2/proposals') > -1:
                        if xpath.find('encryption-algorithm') > -1:
                            old_props['ike-phase-2'][int(prop_id) - 1]['id'] = prop_id
                            old_props['ike-phase-2'][int(prop_id) - 1]['enc'] = change['old_value']
                        if xpath.find('authentication-algorithm') > -1:
                            old_props['ike-phase-2'][int(prop_id) - 1]['id'] = prop_id
                            old_props['ike-phase-2'][int(prop_id) - 1]['auth'] = change['old_value']
                    if xpath.find('local-ts') > -1:
                        del ipsvpn.local_ts
                    if xpath.find('remote-ts') > -1:
                        del ipsvpn.remote_ts
                    continue

                if xpath.find('remote-gw') > -1:
                    gateway = change['new_value']
                    ipsvpn.gateway = gateway
                if xpath.find('local-id') > -1:
                    local_id = change['new_value']
                    if local_id == 'None' or local_id == 'none':
                        del ipsvpn.local_id
                    else:
                        ipsvpn.local_id = local_id
                if xpath.find('remote-id') > -1:
                    remote_id = change['new_value']
                    if remote_id == 'None' or remote_id == 'none':
                        del ipsvpn.remote_id
                    else:
                        ipsvpn.remote_id = remote_id
                if xpath.find('config-mode') > -1:
                    if change['new_value'] == 'pull':
                        ipsvpn.mode_config = ips.MODE_CONFIG.PULL
                    elif change['new_value'] == 'push':
                        ipsvpn.mode_config = ips.MODE_CONFIG.PUSH
                if xpath.find('protocol-mode') > -1:
                    if change['new_value'] == 'tunnel':
                        ipsvpn.child_mode = ips.CHILD_MODE.TUNNEL
                    elif change['new_value'] == 'transport':
                        ipsvpn.child_mode = ips.CHILD_MODE.TRANSPORT
                if xpath.find('ike-version') > -1:
                    ike_version = change['new_value']
                    if ike_version == '1':
                        ipsvpn.ike = ips.IKE.V1
                    elif ike_version == '2':
                        ipsvpn.ike = ips.IKE.V2
                if xpath.find('ike-mode') > -1:
                    ike_mode = change['new_value']
                    if ike_mode == 'main':
                        ipsvpn.ike1_mode = ips.IKE1_MODE.MAIN
                    elif ike_mode == 'aggressive':
                        ipsvpn.ike1_mode = ips.IKE1_MODE.AGGRESIVE
                if xpath.find('ike-phase-1/proposals') > -1:
                    if xpath.find('encryption-algorithm') > -1:
                        props['ike-phase-1'][int(prop_id) - 1]['id'] = prop_id
                        props['ike-phase-1'][int(prop_id) - 1]['enc'] = change['new_value']
                    if xpath.find('authentication-algorithm') > -1:
                        props['ike-phase-1'][int(prop_id) - 1]['id'] = prop_id
                        props['ike-phase-1'][int(prop_id) - 1]['auth'] = change['new_value']
                if xpath.find('ike-phase-2/proposals') > -1:
                    if xpath.find('encryption-algorithm') > -1:
                        props['ike-phase-2'][int(prop_id) - 1]['id'] = prop_id
                        props['ike-phase-2'][int(prop_id) - 1]['enc'] = change['new_value']

                    if xpath.find('authentication-algorithm') > -1:
                        props['ike-phase-2'][int(prop_id) - 1]['id'] = prop_id
                        props['ike-phase-2'][int(prop_id) - 1]['auth'] = change['new_value']

                if xpath.find('ike-phase-1/dh-group') > -1:
                    try:
                        dh_group = int(change['new_value'])
                        if dh_group:
                            ipsvpn.dh_group_add((dh_group), ips.IKE_PHASE.PHASE1)
                    except:
                        pass

                if xpath.find('ike-phase-2/dh-group') > -1:
                    try:
                        dh_group = int(change['new_value'])
                        if dh_group:
                            ipsvpn.dh_group_add((dh_group), ips.IKE_PHASE.PHASE2)
                    except:
                        pass

                if xpath.find('local/rounds') > -1:
                    if xpath.find('authentication-method') > -1:
                        rounds['local'][int(round_id) - 1]['authentication-method'] = change['new_value']
                    if xpath.find('secret') > -1:
                        rounds['local'][int(round_id) - 1]['secret'] = change['new_value']
                    if xpath.find('username') > -1:
                        rounds['local'][int(round_id) - 1]['username'] = change['new_value']
                    if xpath.find('password') > -1:
                        rounds['local'][int(round_id) - 1]['password'] = change['new_value']
                    if xpath.find('ca-cert') > -1:
                        rounds['local'][int(round_id) - 1]['ca-cert'] = change['new_value']
                    if xpath.find('cert') > -1:
                        rounds['local'][int(round_id) - 1]['cert'] = change['new_value']
                    if xpath.find('key-pass') > -1:
                        rounds['local'][int(round_id) - 1]['key-pass'] = change['new_value']
                    if xpath.find('key') > -1 and not xpath.find('key-pass') > -1:
                        rounds['local'][int(round_id) - 1]['key'] = change['new_value']
                    if change['new_value']:
                        rounds['local'][int(round_id) - 1]['modified'] = True
                    rounds['local'][int(round_id) - 1]['id'] = int(round_id)
                if xpath.find('remote/rounds') > -1:
                    if xpath.find('authentication-method') > -1:
                        rounds['remote'][int(round_id) - 1]['authentication-method'] = change['new_value']
                    if xpath.find('secret') > -1:
                        rounds['remote'][int(round_id) - 1]['secret'] = change['new_value']
                    if xpath.find('ca-cert') > -1:
                        rounds['remote'][int(round_id) - 1]['ca-cert'] = change['new_value']
                    if xpath.find('cert') > -1:
                        rounds['remote'][int(round_id) - 1]['cert'] = change['new_value']
                    if xpath.find('key') > -1:
                        rounds['remote'][int(round_id) - 1]['key'] = change['new_value']
                    if xpath.find('key-pass') > -1:
                        rounds['remote'][int(round_id) - 1]['key-pass'] = change['new_value']
                    if change['new_value']:
                        rounds['remote'][int(round_id) - 1]['modified'] = True
                    rounds['remote'][int(round_id) - 1]['id'] = int(round_id)
                if xpath.find('local-ts') > -1:
                    local_ts = IPNetwork(change['new_value'])
                    if local_ts == 'None' or local_ts == 'none':
                        del ipsvpn.local_ts
                    else:
                        ipsvpn.local_ts = local_ts
                if xpath.find('remote-ts') > -1:
                    remote_ts = IPNetwork(change['new_value'])
                    if remote_ts == 'None' or remote_ts == 'none':
                        del ipsvpn.remote_ts
                    else:
                        ipsvpn.remote_ts = remote_ts
                if xpath.find('enable') > -1:
                    if_status = change['new_value']
                    if 'true' in str(if_status).lower():
                        if_status = True
                    else:
                        if_status = False
                if xpath.find('install-virtual-ip') > -1:
                    inst_vip = False if (change['new_value'] is False or change['new_value'] == 'false') else True
                if xpath == 'virtual-ip':
                    vip = change['new_value']

            for key, value in old_props.items():
                for prop in value:
                    if 'enc' in prop and 'auth' in prop and prop['enc']!='None' and prop['auth']!='None':
                        self.logger.debug('Delete proposal: {}'.format(prop))
                        ipsvpn.proposal_del(prop['enc'], prop['auth'], ips.IKE_PHASE.PHASE1 if key == 'ike-phase-1' else ips.IKE_PHASE.PHASE2)

            for key, value in props.items():
                # phase = ips.IKE_PHASE.PHASE1 if key == 'ike-phase-1' else ips.IKE_PHASE.PHASE2
                for prop in value:
                    if 'enc' not in prop and 'auth' not in prop:
                        continue
                    elif 'enc' not in prop:
                        prop['enc'] = self.get_proposal_param(key, name, prop['id'], 'encryption-algorithm')
                        ipsvpn.proposal_del(prop['enc'], prop['auth'], ips.IKE_PHASE.PHASE1 if key == 'ike-phase-1' else ips.IKE_PHASE.PHASE2)
                    elif 'auth' not in prop:
                        prop['auth'] = self.get_proposal_param(key, name, prop['id'], 'authentication-algorithm')
                        ipsvpn.proposal_del(prop['enc'], prop['auth'], ips.IKE_PHASE.PHASE1 if key == 'ike-phase-1' else ips.IKE_PHASE.PHASE2)

                    self.logger.debug('Add proposal {}'.format(prop))
                    ipsvpn.proposal_add(prop['enc'], prop['auth'], ips.IKE_PHASE.PHASE1 if key == 'ike-phase-1' else ips.IKE_PHASE.PHASE2)

            # Set local and remote rounds
            for key, value in rounds.items():
                if key == 'local':
                    for local_round in rounds[key]:
                        if not local_round or 'modified' not in local_round or local_round['modified'] is not True:
                            continue
                        if 'authentication-method' in local_round:
                            if local_round['authentication-method'] == 'pre-shared-keys':
                                ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, int(local_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.PSK, psk=local_round['secret']))
                            elif local_round['authentication-method'] == 'pubkeys':
                                passphrase = None if 'key-pass' not in local_round.keys() else str(local_round['key-pass'])
                                ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, int(local_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.PUBKEY, ca_crt=local_round['ca-cert'], crt=local_round['cert'], key=local_round['key'], passphrase=passphrase))
                            elif local_round['authentication-method'] == 'xauth':
                                ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, int(local_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.XAUTH, user=local_round['username'], psw=local_round['password']))
                            elif local_round['authentication-method'] == 'eap':
                                passphrase = None if 'key-pass' not in local_round.keys() else str(local_round['key-pass'])
                                ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, int(local_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.EAP, user=local_round['username'], psw=local_round['password'], ca_crt=local_round['ca-cert'], crt=local_round['cert'], key=local_round['key'], passphrase=passphrase))
                elif key == 'remote':
                    for remote_round in rounds[key]:
                        if not remote_round or 'modified' not in remote_round or remote_round['modified'] is not True:
                            continue
                        if 'authentication-method' in remote_round:
                            if remote_round['authentication-method'] == 'pre-shared-keys':
                                ipsvpn.auth_add(ips.AUTH_SIDE.REMOTE, int(remote_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.PSK, psk=remote_round['secret']))
                            elif remote_round['authentication-method'] == 'pubkeys':
                                ipsvpn.auth_add(ips.AUTH_SIDE.REMOTE, int(remote_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.PUBKEY, ca_crt=remote_round['ca-cert'], crt=remote_round['cert']))

            if inst_vip is False:
                ipsvpn.vips = None
            else:
                if vip is not None:
                    ipsvpn.vips = IPAddress(vip)

            set_ippass_ipsec(self, prev_status, if_status)

            if prev_status is True:
                ipsecvpn_disconnect(self, ipsvpn)
            if if_status is True:
                ipsecvpn_connect(self, ipsvpn)

def ipsecvpn_for_reenable(self, ifname):
    reenable_ipsecvpn_list = []

    # Proceed only if there is any ipsec vpn
    if len(self.ipsvpn_interfaces) > 0:

        # Get addresses of the adapter, specified by name "ifname"
        adapter_ipv4 = None
        try:
            adapter_ipv4 = self.execute_shell_command("ip addr show {}".format(ifname)).split("inet ")[1].split("/")[0]
        except:
            # On interfaces without inet address exception will be raised, so silently ignore it.
            pass

        adapter_ipv6 = None
        adapter_ipv6_temp = None
        try:
            adapter_ipv6_temp = self.execute_shell_command("ip addr show {} | grep inet6 | grep global ".format(ifname))
            if adapter_ipv6_temp is not None and adapter_ipv6_temp != "":
                adapter_ipv6 = adapter_ipv6_temp.split("inet6 ")[1].split("/")[0].strip('\n').strip()
                if adapter_ipv6.count(':') < 2:
                    adapter_ipv6 = None
        except:
            # On interfaces without inet6 address exception will be raised, so silently ignore it.
            pass

        # If the adapter has at least v4 or v6 address proceed
        if adapter_ipv4 is not None or adapter_ipv6 is not None:
            try:
                adapter_ipv4 = str(adapter_ipv4).encode()
                adapter_ipv6 = str(adapter_ipv6).encode()
            except:
                self.logger.error("Error occurred while encoding IP address for IPSec.")

            # Iterate all ipsec vpn interfaces and search for duplicating IP with that of the adapter, specified by ifname
            for ipsecvpn_iface in self.ipsvpn_interfaces:
                sa = ipsecvpn_iface.get_sa()
                if sa is not None:
                    (ipsecvpn_iface_ip, ipsecvpn_iface_port) = sa['local'].rsplit(':',1)
                    self.logger.debug("adapter_ipv4={}, adapter_ipv6={}, ipsecvpn_iface_ip={}".format(adapter_ipv4, adapter_ipv6, ipsecvpn_iface_ip))
                    if str(adapter_ipv4) == str(ipsecvpn_iface_ip) or str(adapter_ipv6) == str(ipsecvpn_iface_ip):
                        reenable_ipsecvpn_list.append(ipsecvpn_iface)

    return reenable_ipsecvpn_list
