#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import vppapi_if.vppapi_if as vppif
from netaddr import IPNetwork
from sysrepo_interfaces import utils as utils
import ipaddress

TVEAPI_PORT = 7373

def get_vpp_ifname_from_mode(self, ifname):
    vpp_if = vppif.VppApiInterface('sr-interfaces')

    all_ifcfg = self.get_interfaces()
    base_name = ifname
    split_list = str(ifname).split('.')
    if len(split_list) > 1:
        base_name, vlan = split_list

    vpp_name = None
    for key, cfg in all_ifcfg.items():
        if base_name == cfg.get('name',''):
            if cfg.get('mode', '') in ['virtio', 'sr-iov']:
                vpp_name = 'host-' + ifname
            elif 'dpdk' == cfg.get('mode', ''):
                vpp_name = ifname
    del vpp_if
    return vpp_name

def ippass_add_del_management_entries(self, ifname, ip_addr, add):
    if_idx = None
    if_name = None
    if not ip_addr or not ifname:
        return

    vpp_if = vppif.VppApiInterface('ippass_add_del_management_entries')
    if not vpp_if:
        return
        
    (if_name, if_idx) = vpp_if.search_vppif_by_hostifname(ifname)
    if if_name == None or if_idx == None:
        del(vpp_if)
        return

    ifname = get_vpp_ifname_from_mode(self, ifname)
    if_idx = vpp_if.get_ifidx_by_name(ifname)
    self.logger.info('IPPass management: ifname {}, ip_addr {}, if_idx {}, add {}' .format(ifname, ip_addr, if_idx, add))
    host_idx = vpp_if.ippass_get_host_if_idx(if_idx)
    if host_idx is None:
        del(vpp_if)
        return

    ret = vpp_if.ippass_check_status(host_idx=host_idx)
    if ret.retval != 0:
        del(vpp_if)
        return

    # add / del CLI entry:
    ssh_port = utils.get_ssh_port(self)
    if 0 != vpp_if.add_del_ippass_permit_host_entry(host_idx=host_idx, dir_out=0, protocol='tcp', dst_port=ssh_port, ip=ip_addr, add=add):
        self.logger.error('IPPass - failed to apply management rules')
    else:
        self.logger.info('IPPass - {} permit CLI entry' .format('Added' if add else 'Deleted'))

    # add / det TVEAPI entry:
    if 0 != vpp_if.add_del_ippass_permit_host_entry(host_idx=host_idx, dir_out=0, protocol='tcp', dst_port=TVEAPI_PORT, ip=ip_addr, add=add):
        self.logger.error('IPPass - failed to apply management rules')
    else:
        self.logger.info('IPPass - {} permit TVE API entry' .format('Added' if add else 'Deleted'))

    if 0 != vpp_if.add_del_ippass_permit_host_entry(host_idx=host_idx, dir_out=0, protocol='tcp', dst_port=7374, ip=ip_addr, add=add):
        self.logger.error('IPPass - failed to apply management rules')
    else:
        self.logger.info('IPPass - {} permit TVE API entry' .format('Added' if add else 'Deleted'))

    # add / det Keystone auth ports entries:
    if 0 != vpp_if.add_del_ippass_permit_host_entry(host_idx=host_idx, dir_out=0, protocol='tcp', dst_port=5000, ip=ip_addr, add=add):
        self.logger.error('IPPass - failed to apply management rules')
    else:
        self.logger.info('IPPass - {} permit port 5000 entry' .format('Added' if add else 'Deleted'))

    # add / det Keystone auth ports entries:
    if 0 != vpp_if.add_del_ippass_permit_host_entry(host_idx=host_idx, dir_out=0, protocol='tcp', dst_port=35357, ip=ip_addr, add=add):
        self.logger.error('IPPass - failed to apply management rules')
    else:
        self.logger.info('IPPass - {} permit port 35357 entry' .format('Added' if add else 'Deleted'))

    del(vpp_if)

def set_ippass_management(self, name, b_set, b_old, b_new, m_set, m_old, m_new, ip_set, ip_old, ip_new, ip6_set, ip6_old, ip6_new):
    to_del = False
    to_add = False
    apply_addr = ip_new
    apply_addr6 = ip6_new

    if b_set == False and m_set == False and ip_set == False and ip6_set == False:
        return

    # Check if old entry should be deleted from vpp
    if b_set == True or m_set == True:
        if b_new in ['dhcp', 'dhcp6', 'dhcp46'] or m_new == False or m_new == 'false':
            to_del = True
            if ip_set == False or (ip_set == True and ip_new == None):
                apply_addr = ip_old
            if ip6_set == False or (ip6_set == True and ip6_new == None):
                apply_addr6 = ip6_old

    if ip_set == True:
        if ip_new == None or (ip_new != None and ip_new != ip_old):
            to_del = True
            apply_addr = ip_old

    if ip6_set == True:
        if ip6_new == None or (ip6_new != None and ip6_new != ip6_old):
            to_del = True
            apply_addr6 = ip6_old

    # Delete entry
    if to_del == True and apply_addr != None:
        ip = IPNetwork(apply_addr)
        self.logger.info('IPPass - delete old management entry for interface {} address {} '.format(name, str(ip.ip)))
        ippass_add_del_management_entries(self, name, str(ip.ip), 0)

    if to_del == True and apply_addr6 != None:
        ip = IPNetwork(apply_addr6)
        self.logger.info('IPPass - delete old management entry for interface {} address {} '.format(name, str(ip.ip)))
        ippass_add_del_management_entries(self, name, str(ip.ip), 0)

    # Check if new entry should be installed
    if b_set == True or m_set == True:
        if (b_new != None and b_new not in ['dhcp', 'dhcp6', 'dhcp46']) or 'true' in str(m_new).lower():
            to_add = True
            apply_addr = ip_new
            if ip_set == False:
                apply_addr = ip_old

            apply_addr6 = ip6_new
            if ip6_set == False:
                apply_addr6 = ip6_old

    else:
        if ip_set == True and ip_new != None:
            if 'true' in str(m_old).lower() and (b_old != None and b_old not in ['dhcp', 'dhcp6', 'dhcp46']):
                to_add = True
                apply_addr = ip_new
        if ip6_set == True and ip6_new != None:
            if 'true' in str(m_old).lower() and (b_old != None and b_old not in ['dhcp', 'dhcp6', 'dhcp46']):
                to_add = True
                apply_addr6 = ip6_new

    # Add entry
    if to_add == True:
        if ip_new == None:
            apply_addr = ip_old
        if apply_addr != None:
            ip = IPNetwork(apply_addr)
            self.logger.info('IPPass - add new entry management for interface {} address {} '.format(name, str(ip.ip)))
            ippass_add_del_management_entries(self, name, str(ip.ip), 1)
        if ip6_new == None:
            apply_addr6 = ip6_old
        if apply_addr6 != None:
            ip = IPNetwork(apply_addr6)
            self.logger.info('IPPass - add new entry management for interface {} address {} '.format(name, str(ip.ip)))
            ippass_add_del_management_entries(self, name, str(ip.ip), 1)

def config_interface_ippass(self, name, config, change_set):
    if config is None or utils.vpp_not_running():
        return
    ippass_en = config.get('ip-passthrough', False)

    if False == ippass_en:
        return

    self.logger.info("On interface {} ippass is enabled".format(name))

    # if config and 'enable' in config:
    # old_enable_nat = config['enable']

    vpp_if = vppif.VppApiInterface('sr-interfaces')
    vpp_if_id = None
    host_if_id = None

    if change_set:
        vpp_if_id = vpp_if.get_ifidx_by_name(get_vpp_ifname_from_mode(self, name))

        if vpp_if_id is None:
            del(vpp_if)
            return

        for xpath, change in change_set.items():
            vpp_if_name = vpp_if.get_ifname_by_ifidx(vpp_if_id)
            vpp_if_id = vpp_if.get_ifidx_by_name(vpp_if_name)
            host_if_id = vpp_if.ippass_get_host_if_idx(vpp_if_id)
            if host_if_id is None:
                del(vpp_if)
                return

        b_set = False
        bproto = None
        b_old = self.get_interface_param(name, 'bootproto')

        m_set = False
        management = None
        m_old = self.get_interface_param(name, 'management')

        ipv4_set = False
        ipv4_addr_mask = None
        ipv4_old = self.get_interface_param(name, 'ipv4-prefix')

        ipv6_set = False
        ipv6_addr_mask = None
        ipv6_old = self.get_interface_param(name, 'ipv6-prefix')

        for xpath, change in change_set.items():
            if xpath.find('bootproto') > -1:
                b_set = True
                bproto = change.get("new_value", None)
                if change.get("old_value","") in ['dhcp', 'dhcp6', 'dhcp46'] and change.get("new_value","") not in ['dhcp', 'dhcp6', 'dhcp46']:
                    vpp_if.add_del_ippass_permit_host_entry(host_if_id, 'dhcp', add=0)

                if change.get("new_value","") in ['dhcp', 'dhcp6', 'dhcp46'] and change.get("old_value","") not in ['dhcp', 'dhcp6', 'dhcp46']:
                    vpp_if.add_del_ippass_permit_host_entry(host_if_id, 'dhcp')
            if xpath.find('management') > -1:
                m_set = True
                management = change.get("new_value", None)
            if xpath.find('ipv4-prefix') > -1:
                ipv4_set = True
                if change.get('operation', None) != 'DELETED':
                    ipv4_addr_mask = change.get("new_value", None)
            if xpath.find('ipv6-prefix') > -1:
                ipv6_set = True
                if change.get('operation', None) != 'DELETED':
                    ipv6_addr_mask = change.get("new_value", None)

        del(vpp_if)
        set_ippass_management(self, name, b_set, b_old, bproto, m_set, m_old, management,
                                     ipv4_set, ipv4_old, ipv4_addr_mask, ipv6_set, ipv6_old, ipv6_addr_mask)

    else:
        self.logger.info("IPPass apply startup config")
        if config.get('mode', 'virtio') == 'dpdk':
            vpp_if_id = vpp_if.get_ifidx_by_name(name)
        else:
            vpp_if_id = vpp_if.get_ifidx_by_name("host-{}".format(name))

        if vpp_if_id is None:
            del(vpp_if)
            return

        host_if_id = vpp_if.ippass_get_host_if_idx(vpp_if_id)
        if host_if_id is None:
            del(vpp_if)
            return

        name = name.split('.')[0]

        if config.get('bootproto', '') in ['dhcp', 'dhcp6', 'dhcp46']:
            self.logger.info('Enable dhcp on ippass host inteface')
            vpp_if.add_del_ippass_permit_host_entry(host_if_id, 'dhcp')

        del(vpp_if)
        if 'true' in str(config.get('management', False)).lower():
            if config.get('bootproto', '') == 'none':
                ipv4_addr_mask = config.get('ipv4-prefix', '')
                if ipv4_addr_mask != '':
                    ip = IPNetwork(ipv4_addr_mask)
                    ippass_add_del_management_entries(self, name, str(ip.ip), 1)

                ipv6_addr_mask = config.get('ipv6-prefix', '')
                if ipv6_addr_mask != '':
                    ip = IPNetwork(ipv6_addr_mask)
                    ippass_add_del_management_entries(self, name, str(ip.ip), 1)

def handle_vpp_interface_addresses(self, ifname, ipv4_prefix=None, ipv6_prefix=None):
    """ ifname param - br-dataX
    """

    if self.get_network_mode() != 'vpp':
        return

    bvi_idx = None
    host_if = None
    host_idx = None

    self.logger.info("handle_vpp_interface_addresses - {} - {} - {}" .format(ifname, ipv4_prefix, ipv6_prefix))
    vpp_if = vppif.VppApiInterface('if-addresses')
    host_if = vpp_if.get_tap_interface_by_host_ifname(ifname)
    if not host_if:
        del(vpp_if)
        return
    bridge = vpp_if.get_interface_bd_id(host_if['sw_if_index'])
    if bridge:
        bvi_idx = vpp_if.get_bd_bvi_by_bdidx(bridge)
    host_idx = host_if['sw_if_index']

    if not bvi_idx or not host_idx:
        del(vpp_if)
        return

    if ipv4_prefix:
        ipv4_a = ipaddress.ip_interface(ipv4_prefix)
        ret = vpp_if.add_interface_address(idx=bvi_idx,addr=ipv4_prefix)
        if ret:
            self.logger.error(" Error adding interface {} address {} in vpp - {} ". format(ifname,ipv4_prefix,ret))
            del(vpp_if)
            return
        ret = vpp_if.drctl_set_ip( host_idx, ipv4_a.ip.packed )
        self.logger.info("handle_vpp_interface_addresses {} ipv4 - {} added" .format(ifname, ipv4_a.packed))

    if ipv6_prefix:
        ipv6_a = ipaddress.ip_interface(ipv6_prefix)
        ret = vpp_if.add_interface_address(idx=bvi_idx,addr=ipv6_prefix)
        if ret:
            self.logger.error(" Error adding interface {} address {} in vpp - {} ". format(ifname,ipv6_prefix,ret))
            del(vpp_if)
            return
        ret = vpp_if.drctl_set_ip( host_idx, ipv6_a.ip.packed )
        vpp_if.ip6_ra_config(bvi_idx, suppress=1)
        self.logger.info("handle_vpp_interface_addresses {} ipv6 - {} added" .format(ifname, ipv6_a.ip.packed))

    del(vpp_if)
    return

def set_ippass_enable_udp_dst_port(self, port, enable):
    if_cfg = self.get_interfaces()
    ippass_ifs = list(filter(lambda x: True if (x.get('type', '') in ['ethernet', 'vlan', 'lte', 'wifi', 'pppoe']) and x.get('ip-passthrough', False) else False, list(if_cfg.values())))

    if not ippass_ifs or utils.vpp_not_running():
        return

    vpp_if = vppif.VppApiInterface('sr-interfaces')
    for interface in ippass_ifs:
        ifname = interface.get('name', 'noname')

        (if_name, vpp_idx) = vpp_if.search_vppif_by_hostifname(ifname)
        if if_name == None or vpp_idx == None:
            continue

        ippass_host_idx = vpp_if.ippass_get_host_if_idx(vpp_idx)
        ret = vpp_if.add_del_ippass_permit_host_entry(ippass_host_idx, 'udp', dst_port=int(port),  add=1 if enable else 0)
        self.logger.debug('add_del_ippass_permit_host_entry(ippass_host_idx {}, \'udp\', dst_port={},  add={}, ret {})'.format(ippass_host_idx, port, 1 if enable else 0, ret))
    del(vpp_if)

def set_ippass_enable_esp_proto(self, enable):
    if_cfg = self.get_interfaces()
    ippass_ifs = list(filter(lambda x: True if (x.get('type', '') in ['ethernet', 'vlan', 'lte', 'wifi', 'pppoe'] ) and x.get('ip-passthrough', False) else False, list(if_cfg.values())))

    if not ippass_ifs or utils.vpp_not_running():
        return

    vpp_if = vppif.VppApiInterface('sr-interfaces')
    for interface in ippass_ifs:
        (if_name, vpp_idx) = vpp_if.search_vppif_by_hostifname(interface.get('name', 'noname'))
        if if_name == None or vpp_idx == None:
            continue

        ippass_host_idx = vpp_if.ippass_get_host_if_idx(vpp_idx)
        ret = vpp_if.add_del_ippass_permit_host_entry(ippass_host_idx, 'esp', add=1 if enable else 0)
        self.logger.debug('add_del_ippass_permit_host_entry(ippass_host_idx {}, \'esp\', add={}, ret {})'.format(ippass_host_idx, 1 if enable else 0, ret))
    del(vpp_if)

def interface_to_vsw_vpp(self, ifname, ifmode, vsw, apply_def=False):
    """ ifname: dataXX.vlan if vlan is configured
        vsw:    br-dataXX.vlan if vlan is configured
    """
    # if apply_def == False and 'br-'+ifname == vsw:
    #     return

    vpp_if = vppif.VppApiInterface('vsw-startup')
    vpp_iface = "host-" + ifname if ifmode == 'virtio' else ifname
    ifidx = vpp_if.get_ifidx_by_name(vpp_iface)
    bridx = vpp_if.get_bd_id_by_tag(vsw)
    if not bridx:
        if apply_def:
            ret = vpp_if.create_bridge_domain(ifidx, tag=vsw)
            bridx = vpp_if.get_bd_id_by_tag(vsw)
        else:
            del(vpp_if)
            return 1
    if ifidx and bridx:
        vpp_if.add_to_bridge(bridx,ifidx)
        self.logger.debug("Divert interface {} into BD with tag {}" .format(ifname, vsw))
    del(vpp_if)
    return 0
